"""JSON parser. This program should parse JSON into Python.

The rather peculiar title refers to the way of building JSON by moving through
a tree: https://www.json.org/json-en.html
"""

from typing import Any, Tuple, Union


DEBUG = False


def handle_value(input_string: str) -> Any:
    """Routing method depending on the value."""

    input_string = input_string.strip()

    if len(input_string) > 0:

        if input_string[0] == '"' and input_string[-1] == '"':
            return handle_string(input_string)

        if input_string[0] == '[' and input_string[-1] == ']':
            return handle_array(input_string)

        if input_string[0] == '{' and input_string[-1] == '}':
            return handle_object(input_string)

        if input_string == "true":
            return True

        if input_string == "false":
            return False

        if input_string == "null":
            return None

        # if it isn't any of the above, it must be a number
        return handle_number(input_string)

    return ''


def handle_string(input_string: str) -> str:
    """Tree arrives on a string, handle it.

    "string"
    """

    input_string = input_string.strip()

    return input_string[1:-1]


def handle_number(input_string: str) -> Union[int, float]:
    """Tree arrives on a number, handle it.

    number
    """

    input_string = input_string.strip()

    try:
        integer: int = int(input_string)

        return integer

    except ValueError:
        fraction: float = float(input_string)

        return fraction


def handle_array(string: str) -> list:
    """Tree arrives on an array, handle it."""

    # prepare empty list for result array
    output: list = list()

    # take away the brackets, we're only working on what's in the array
    input_string: str = string[1:-1].strip()

    debug_input_string(input_string)

    start = 0
    end = 0

    # As long as end_index hasn't reached end of the input_string, continue.
    while end < len(input_string):

        # Update start and end to select a sub-element.
        start, end = set_selection(input_string, start, end)

        # Process selection

        current_value: str = ""

        # If we get comma at end of sub-string, we need to append it to output.
        if input_string[end] == ',':
            current_value = input_string[start:end]
            start = end + 1

        # If we reach end of sub-string, we need to append it to output.
        elif (end == len(input_string) - 1) \
             or \
             (input_string[start] == '[' and \
              input_string[end] == ']') \
             or \
             (input_string[start] == '{' and \
              input_string[end] == '}'):
            current_value = input_string[start:end + 1]
            start = end + 1

        # Making sure we don't have an empty element.
        if len(current_value) > 0:
            value: Any = handle_value(current_value.strip())
            debug(f"\n   current_value: {current_value} -- value: {value} -- {type(value)}\n")

            output.append(value)

        end += 1

    return output


def handle_object(string: str) -> dict:
    """Tree arrives on an object, handle it."""

    # prepare empty dict for result object
    output: dict = dict()

    # take away the curly brackets, we're only working on what's in the object
    input_string: str = string[1:-1].strip()

    debug_input_string(input_string)

    start = 0
    end = 0

    # As long as end_index hasn't reached end of the input_string, continue.
    while end < len(input_string):

        # Update start and end to select a sub-element.
        start, end = set_selection(input_string, start, end)

        # Process selection

        current_key: str = ""
        current_value: str = ""

        # If we get a ':' at end of sub-string, we have a key for the dict.
        if input_string[end] == ':': 
            current_key = input_string[start:end]
            start = end + 1

        # If we get comma at end of sub-string, we have a value for the dict.
        elif input_string[end] == ',':
            current_value = input_string[start:end]
            start = end + 1

        # If we reach end of sub-string, we have a value for the dict.
        elif (end == len(input_string) - 1) \
             or \
             (input_string[start] == '[' and \
              input_string[end] == ']') \
             or \
             (input_string[start] == '{' and \
              input_string[end] == '}'):
            current_value = input_string[start:end + 1]
            start = end + 1

        # Making sure we don't have empty elements.
        if len(current_key) > 0:
            key: Any = handle_value(current_key.strip())
            debug(f"\n   current_key: {current_key} -- key: {key} -- {type(key)}")

        if len(current_value) > 0:
            value: Any = handle_value(current_value.strip())
            debug(f"\n   current_value: {current_value} -- value: {value} -- {type(value)}")

            output[key] = value

        end += 1

    return output


def move_to_end(end_index: int, input_string: str, closing_target: str) -> int:
    """Move cursor to the end of sub-array or sub-object.
    
    Return updated end_index.
    """

    if closing_target == ']':
        opening_target = '['
    elif closing_target == '}':
        opening_target = '{'

    # Anytime a new opening bracket is found, we need to ignore the matching
    # closing bracket.
    closing = 1
    while input_string[end_index] != closing_target or closing >= 1:
        end_index += 1
        if input_string[end_index] == opening_target:
            closing += 1
        elif input_string[end_index] == closing_target:
            closing -= 1

    return end_index


def set_selection(input_string: str, start: int, end: int) -> Tuple[int]:
    """Update start and end to correct position for a selection to process.
    
    Return updated indexes.
    """

    # We get space or comma at start of sub-string, move past it.
    if input_string[start] in (' ', ','):
        start += 1
        end = start

    # We get sub-array, put end_index to end of sub-array
    elif input_string[start] == '[':
        end = move_to_end(end, input_string, ']')

    # We get sub-object, put end_index to end of sub-object.
    elif input_string[start] == '{':
        end = move_to_end(end, input_string, '}')

    return (start, end)


def debug_input_string(input_string):
    """Debugging function for visualizing the input string."""

    debug()
    debug(' ', end=' ')
    for i in range(len(input_string)):
        debug(input_string[i], ' ', end='')
    debug()
    for i in range(len(input_string)):
        debug("%3d" % i, end='')
    debug()


def debug(*message, **args):
    if DEBUG:
        print(*message, **args)

