"""Test cases for the parser"""

import unittest
import monkey


class TestValue(unittest.TestCase):
    """Going through a JSON value."""

    def test_string(self):
        """String (always in double quotes)"""

        json_input = '"string"'
        result = monkey.handle_value(json_input)
        self.assertEqual(result, "string")
        self.assertEqual(len(result), 6)

    def test_number(self):
        """Number"""

        json_input = "65"
        result = monkey.handle_value(json_input)
        self.assertEqual(result, 65)
        self.assertIsInstance(result, int)

    def test_true(self):
        """true"""

        json_input = "true"
        result = monkey.handle_value(json_input)
        self.assertEqual(result, True)
        self.assertTrue(result)

    def test_false(self):
        """false"""

        json_input = "false"
        result = monkey.handle_value(json_input)
        self.assertEqual(result, False)
        self.assertFalse(result)

    def test_null(self):
        """null"""

        json_input = "null"
        result = monkey.handle_value(json_input)
        self.assertEqual(result, None)
        self.assertIsNone(result)

    def test_object_with_string(self):
        """Object with string"""

        json_input = '{ "string": "value" }'
        result = monkey.handle_value(json_input)
        self.assertEqual(result, {"string": 'value'})
        self.assertEqual(result, {'string': "value"})
        self.assertEqual(result, dict(string="value"))
        self.assertIsInstance(result, dict)

    def test_object_with_number(self):
        """Object with number"""

        json_input = '{ "string": 53 }'
        result = monkey.handle_value(json_input)
        self.assertEqual(result, {"string": 53})
        self.assertEqual(result, dict(string=53))
        self.assertIsInstance(result, dict)

    def test_simple_array(self):
        """Simple array"""

        json_input = '[ "string" ]'
        result = monkey.handle_value(json_input)
        self.assertEqual(result, monkey.handle_array(json_input))
        self.assertEqual(result, ["string"])
        self.assertIsInstance(result, list)

    def test_nested_array(self):
        """Nested array"""

        json_input = '[ "string", 55, "other string", false ]'
        result = monkey.handle_value(json_input)
        self.assertEqual(result, ["string", 55, "other string", False])
        self.assertIsInstance(result, list)


class TestString(unittest.TestCase):
    """Going through a JSON string."""

    def test_empty_string(self):
        """Empty string"""

        json_input = '""'
        result = monkey.handle_string(json_input)
        self.assertEqual(result, '')
        self.assertEqual(len(result), 0)

    def test_simple_string(self):
        """Simple string"""

        json_input = '"string"'
        result = monkey.handle_string(json_input)
        self.assertEqual(result, "string")
        self.assertEqual(len(result), 6)

    def test_string_with_spaces(self):
        """String with spaces"""

        json_input = '"string with spaces"'
        result = monkey.handle_string(json_input)
        self.assertEqual(result, "string with spaces")
        self.assertEqual(len(result), 18)

    def test_string_with_quotation_marks(self):
        """String with quotation marks"""

        json_input = '"\"string with quotation marks\""'
        result = monkey.handle_string(json_input)
        self.assertEqual(result, '"string with quotation marks"')
        self.assertEqual(len(result), 29)


class TestNumber(unittest.TestCase):
    """Going through a JSON number."""

    def test_integer(self):
        """Integer"""

        json_input = "4"
        self.assertEqual(monkey.handle_number(json_input), 4)

        json_input = "13"
        self.assertEqual(monkey.handle_number(json_input), 13)

        json_input = "02"
        self.assertEqual(monkey.handle_number(json_input), 2)

        json_input = "1052"
        self.assertEqual(monkey.handle_number(json_input), 1052)

        json_input = "40840"
        self.assertEqual(monkey.handle_number(json_input), 40840)

    def test_negative_integer(self):
        """Negative integer"""

        json_input = "-078"
        self.assertEqual(monkey.handle_number(json_input), -78)

        json_input = "-4"
        self.assertEqual(monkey.handle_number(json_input), -4)

        json_input = "-102"
        self.assertEqual(monkey.handle_number(json_input), -102)

        json_input = "-5509500"
        self.assertEqual(monkey.handle_number(json_input), -5509500)

    def test_fraction(self):
        """Fraction"""

        json_input = "-5.87"
        self.assertEqual(monkey.handle_number(json_input), -5.87)

        json_input = "-0.54"
        self.assertEqual(monkey.handle_number(json_input), -0.54)

        json_input = "66.8970"
        self.assertEqual(monkey.handle_number(json_input), 66.897)

        json_input = "07.040"
        self.assertEqual(monkey.handle_number(json_input), 7.04)


class TestArray(unittest.TestCase):
    """Going through a JSON array."""

    def test_empty_array(self):
        """Empty array"""

        json_input = "[ ]"
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [])
        self.assertEqual(result, list())
        self.assertIsInstance(result, list)

    def test_array_with_empty_string(self):
        """Array with empty string"""

        json_input = '[ "" ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [""])
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], str)

    def test_array_with_one_value_string(self):
        """Array with one value: string"""

        json_input = '[ "value as a string" ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, ['value as a string'])
        self.assertEqual(result, ["value as a string"])
        self.assertIsInstance(result, list)

    def test_array_with_one_value_number(self):
        """Array with one value: number"""

        json_input = '[ 23 ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [23])
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result, list)

    def test_array_with_one_value_empty_object(self):
        """Array with one value: empty object"""

        json_input = '[ { } ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [{}])
        self.assertEqual(result, [dict()])
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], dict)

    def test_array_with_one_value_object(self):
        """Array with one value: object"""

        json_input = '[ { "key" : 23 } ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [{'key': 23}])
        self.assertEqual(result, [dict(key=23)])
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], dict)

    def test_array_with_one_value_empty_array(self):
        """Array with one value: empty array"""

        json_input = '[ [ ] ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [[]])
        self.assertEqual(result, [list()])
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], list)

    def test_array_with_one_value_array(self):
        """Array with one value: array"""

        json_input = '[ [ "string" ] ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [["string"]])
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], list)
        self.assertIsInstance(result[0][0], str)

    def test_array_with_string_and_array(self):
        """Array with string and array"""

        json_input = '[ "string", [ "other string" ] ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, ["string", ["other string"]])
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], str)
        self.assertIsInstance(result[1], list)

    def test_array_with_one_value_true(self):
        """Array with one value: true"""

        json_input = '[ true ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [True])
        self.assertIsInstance(result, list)
        self.assertTrue(result[0])

    def test_array_with_one_value_false(self):
        """Array with one value: false"""

        json_input = '[ false ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [False])
        self.assertIsInstance(result, list)
        self.assertFalse(result[0])

    def test_array_with_one_value_null(self):
        """Array with one value: null"""

        json_input = '[ null ]'
        result = monkey.handle_array(json_input)
        self.assertEqual(result, [None])
        self.assertIsInstance(result, list)
        self.assertIsNone(result[0])

    def test_array_list_in_list(self):
        """List embedded in another list."""

        json_input = (
            '[ 34, "string", [ "other list", [null, 33], true]]'
        )
        result = monkey.handle_array(json_input)
        self.assertEqual(
            result,
            [34, "string", ["other list", [None, 33], True]]
        )
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[2], list)
        self.assertEqual(len(result), 3)

    def test_array_with_more_combining_values_end_with_dict(self):
        """Array with more combining values, end with a dict."""

        json_input = (
            '[ null, "string with spaces", 56, true, [23, 45], { "key": 5 } ]'
        )
        result = monkey.handle_array(json_input)
        self.assertEqual(
            result,
            [None, "string with spaces", 56, True, [23, 45], {"key": 5}]
        )
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 6)

    def test_array_more_embedding(self):
        """Array with more combining values, more embedding, end with null."""

        json_input = (
            '[ null, {"string with spaces": "value", "key": 23 }, 5, true,'
            '[23, ["string", "other string"], 45], null ]'
        )
        result = monkey.handle_array(json_input)
        self.assertEqual(
            result,
            [None, {"string with spaces": "value", "key": 23}, 5, True,
             [23, ["string", "other string"], 45], None]
        )
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 6)


class TestObject(unittest.TestCase):
    """Going through a JSON object."""

    def test_empty_object(self):
        """Empty object"""

        json_input = "{ }"
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {})
        self.assertEqual(result, dict())
        self.assertIsInstance(result, dict)

    def test_object_with_one_value_string(self):
        """Object with one value: string"""

        json_input = '{ "string" : "value as a string" }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": "value as a string"})
        self.assertEqual(result, dict(string="value as a string"))
        self.assertIsInstance(result, dict)

    def test_object_with_one_value_number(self):
        """Object with one value: number"""

        json_input = '{ "string" : 23 }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": 23})
        self.assertEqual(result, dict(string=23))
        self.assertIsInstance(result, dict)

    def test_object_with_one_value_object(self):
        """Object with one value: object"""

        json_input = '{ "string" : { } }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": {}})
        self.assertEqual(result, {'string': dict()})
        self.assertEqual(result, dict(string={}))
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['string'], dict)

    def test_object_with_one_value_array(self):
        """Object with one value: array"""

        json_input = '{ "string" : [ ] }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": []})
        self.assertEqual(result, {'string': list()})
        self.assertEqual(result, dict(string=[]))
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['string'], list)

    def test_object_with_one_value_true(self):
        """Object with one value: true"""

        json_input = '{ "string" : true }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": True})
        self.assertEqual(result, dict(string=True))
        self.assertIsInstance(result, dict)
        self.assertTrue(result['string'])

    def test_object_with_one_value_false(self):
        """Object with one value: false"""

        json_input = '{ "string" : false }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": False})
        self.assertEqual(result, dict(string=False))
        self.assertIsInstance(result, dict)
        self.assertFalse(result['string'])

    def test_object_with_one_value_null(self):
        """Object with one value: null"""

        json_input = '{ "string" : null }'
        result = monkey.handle_object(json_input)
        self.assertEqual(result, {"string": None})
        self.assertEqual(result, dict(string=None))
        self.assertIsInstance(result, dict)
        self.assertIsNone(result['string'])

    def test_object_in_object(self):
        """Object embedded in another object."""

        json_input = '{ "string": { "key": 23 } }'
        result = monkey.handle_object(json_input)
        self.assertEqual(
            result,
            {'string': {'key': 23}}
        )
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['string'], dict)
        self.assertIsInstance(result['string']['key'], int)

    def test_object_with_more_combining_values_end_with_object(self):
        """Object with more combining values, end with an object."""

        json_input = (
            '{ 33: null, "apple": {"key": [ 23, "abc" ] } }'
        )
        result = monkey.handle_object(json_input)
        self.assertEqual(
            result,
            {33: None, 'apple': {'key': [23, 'abc']}}
        )
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['apple'], dict)
        self.assertIsInstance(result['apple']['key'], list)

    def test_object_with_more_combining_values_dont_end_with_object(self):
        """Object with more combining values, don't end with an object."""

        json_input = (
            '{ 33: null, "apple": {"key": [ 23, "abc" ] }, "other": 22 }'
        )
        result = monkey.handle_object(json_input)
        self.assertEqual(
            result,
            {33: None, 'apple': {'key': [23, 'abc']}, 'other': 22}
        )
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['apple'], dict)
        self.assertIsInstance(result['apple']['key'], list)

    def test_object_with_array_as_key_dont_end_with_object(self):
        """Object with array as a key, don't end with an object."""

        json_input = (
            '{ 33: null, "apple": [ 23, "abc" ], "other": 22 }'
        )
        result = monkey.handle_object(json_input)
        self.assertEqual(
            result,
            {33: None, 'apple': [23, 'abc'], 'other': 22}
        )
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result['apple'], list)
        self.assertEqual(len(result['apple']), 2)

    def test_object_more_embedding(self):
        """Object with more combining values, more embedding, end with null."""

        json_input = (
            '{ "foo": "bar", "key": {34: 45, "apple": "orange"}, "other": ['
            'null, 15, false], 32: true, "a": [ {88: "ahha"}, false] }'
        )
        result = monkey.handle_object(json_input)
        self.assertEqual(
            result,
            {"foo": "bar", "key": {34: 45, "apple": "orange"}, "other": [
            None, 15, False], 32: True, "a": [{88: "ahha"}, False]}
        )
        self.assertIsInstance(result, dict)
        self.assertIsInstance(result["key"], dict)
        self.assertIsInstance(result["other"], list)
        self.assertIsNone(result["other"][0])
        self.assertTrue(result[32])
        self.assertIsInstance(result["a"][0], dict)


class TestIndentedObject(unittest.TestCase):
    """Rewrite of TestObject but with indentation."""

    def test_object_with_one_value_string(self):
        """Object with one value: string"""

        json_input = '{"string" : "value as a string"}'
        json_input2 = '{\n\t"string" : "value as a string"\n}'
        json_input3 = (
            '{'
            '   "string" : "value as a string"'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        result3 = monkey.handle_object(json_input3)
        self.assertEqual(result, {'string': 'value as a string'})
        self.assertEqual(result, result2)
        self.assertEqual(result2, result3)

    def test_object_with_one_value_number(self):
        """Object with one value: number"""

        json_input = '{ "string" : 23 }'
        json_input2 = '{\n\t"string" : 23\n}'
        json_input3 = (
            '{'
            '   "string" : 23'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        result3 = monkey.handle_object(json_input3)
        self.assertEqual(result, {"string": 23})
        self.assertEqual(result, result2)
        self.assertEqual(result2, result3)

    def test_object_with_one_value_object(self):
        """Object with one value: object"""

        json_input = '{ "string" : { } }'
        json_input2 = '{\n\t"string" : {\n\t}\n}'
        json_input3 = (
            '{'
            '   "string" : {'
            '   }'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        result3 = monkey.handle_object(json_input3)
        self.assertEqual(result, {"string": {}})
        self.assertEqual(result, result2)
        self.assertEqual(result2, result3)

    def test_object_with_one_value_array(self):
        """Object with one value: array"""

        json_input = '{ "string" : [ ] }'
        json_input2 = '{\n\t"string" : [\n\t]\n}'
        json_input3 = (
            '{'
            '   "string" : ['
            '   ]'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        result3 = monkey.handle_object(json_input3)
        self.assertEqual(result, {"string": []})
        self.assertEqual(result, result2)
        self.assertEqual(result2, result3)

    def test_object_in_object(self):
        """Object embedded in another object."""

        json_input = '{ "string": { "key": 23 } }'
        json_input2 = '{\n\t"string": {\n\t\t"key": 23\n\t}\n}'
        json_input3 = (
            '{'
            '   "string": {'
            '       "key": 23'
            '   }'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        result3 = monkey.handle_object(json_input3)
        self.assertEqual(result, {'string': {'key': 23}})
        self.assertEqual(result, result2)
        self.assertEqual(result2, result3)

    def test_object_with_more_combining_values_end_with_object(self):
        """Object with more combining values, end with an object."""

        json_input = '{ 33: null, "apple": {"key": [ 23, "abc" ] } }'
        json_input2 = (
            '{\n\t33: null,\n\t"apple": {\n\t\t"key": [ 23, "abc" ]\n\t}\n}'
        )
        json_input3 = (
            '{'
            '   33: null,'
            '   "apple": {'
            '       "key": [23, "abc"]'
            '   }'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        result3 = monkey.handle_object(json_input3)
        self.assertEqual(result, {33: None, 'apple': {'key': [23, 'abc']}})
        self.assertEqual(result, result2)
        self.assertEqual(result2, result3)

    def test_object_with_more_combining_values_dont_end_with_object(self):
        """Object with more combining values, don't end with an object."""

        json_input = (
            '{ 33: null, "apple": {"key": [ 23, "abc" ] }, "other": 22 }'
        )
        json_input2 = (
            '{'
            '   33: null,'
            '   "apple": {'
            '       "key": [ 23, "abc" ]'
            '   },'
            '   "other": 22,'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        self.assertEqual(
            result,
            {33: None, 'apple': {'key': [23, 'abc']}, 'other': 22}
        )
        self.assertEqual(result, result2)

    def test_object_more_embedding(self):
        """Object with more combining values, more embedding, end with null."""

        json_input = (
            '{ "foo": "bar", "key": {34: 45, "apple": "orange"}, "other": ['
            'null, 15, false], 32: true, "a": [ {88: "ahha"}, false] }'
        )
        json_input2 = (
            '{'
            '   "foo": "bar",'
            '   "key": {34: 45, "apple": "orange"},'
            '   "other": ['
            '       null, 15, false'
            '   ],'
            '   32: true,'
            '   "a": ['
            '       {88: "ahha"},'
            '       false'
            '   ]'
            '}'
        )
        result = monkey.handle_object(json_input)
        result2 = monkey.handle_object(json_input2)
        self.assertEqual(
            result,
            {"foo": "bar", "key": {34: 45, "apple": "orange"}, "other": [
            None, 15, False], 32: True, "a": [{88: "ahha"}, False]}
        )
        self.assertEqual(result, result2)


if __name__ == '__main__':
    unittest.main()

